from __future__ import annotations
from types import LambdaType, NoneType
from typing import TYPE_CHECKING, Dict, Union, List, ItemsView, Any, Type, Callable, cast
from ..savable import Savable
from .constraint import Constraint
from ..resource import Resource
from ..edata.edata import EData
from ...parser.term import Term
# from ...parser.leaf import Model, QueryingProxy, Domain, ColumnProxy
from ...parser.inspect_parser.model import Model
from ...parser.inspect_parser.domain import Domain
from ..db.table_column import TableColumn
from collections import OrderedDict
from ..edata.proto import Proto
from uuid import UUID
# from ..edata.elist import ElistMetacls
from ...parser.inspect_parser.inspect_parser import InspectParser
from ...parser.proxy_base_cls import ColumnProxyBase
from copy import deepcopy
from ...resource.join_resource.join_resource import JoinResource

from ...errors import DeserializeError
from ...epure import escript

if TYPE_CHECKING:
    from .table_storage import TableStorage
    from ...parser.term_parser import TermParser

from .db_entity import DbEntity
from ...helpers.type_helper import check_type
from ...errors import EpureError, DbError
from .table_header import TableHeader
from ..data_promise import FieldPromise, DataPromise, ElistPromise, EsetPromise
from ...epure import Epure

class Table(DbEntity):
    header:TableHeader
    if TYPE_CHECKING:
        parser: TermParser
    # querying_proxy: QueryingProxy
    # resource_proxy: Domain
    querying_proxy:Model
    resource_proxy:Domain

    if TYPE_CHECKING:
        resource:TableStorage

    @property
    def db(self):
        return self.resource

    def __init__(self, name: str,
            header:Union[TableHeader, Dict[str, Any]]=None, resource:Resource=None, namespace:str = '', parser=InspectParser) -> None:        
        check_type('header', header, [TableHeader, dict, NoneType])
        from ...parser.term_parser import TermParser

        self._set_header(header)
        super().__init__(name, namespace, resource)

        # self.parser = TermParser(self)
        if parser == TermParser:
            parser = TermParser(self)
        
        self.parser = parser
        self.querying_proxy = Model(self.db, self)
        self.resource_proxy = Domain(self.db)
        


    # def _serialize(self, node: Savable) -> Dict[str, str]:
    #     res = {}
    #     for field_name, field_type in node.annotations.items():
    #         if isinstance(field_type, Constraint):
    #             field_type = field_type.py_type
                
    #         if node.is_excluded(field_name, field_type):
    #             continue
    #         if field_name not in self.header:
    #             continue
    #         if not hasattr(node, field_name):
    #             continue

    #         field_val = getattr(edata, field_name, None)
            
    #         field_val = self._serialize_field_val(field_val, field_type)

    #         res[field_name] = field_val
        
    #     return res

    @classmethod
    def is_excluded(self, edata, atr_name:str, type_hint:Any='') -> bool:
        if type_hint in (NoneType, None):
            return True
        return super().is_excluded(edata, atr_name, atr_name)


    def _serialize_field_val_to_sql(self, field_val, field_type=None, field_name=None, _rec_depth=None, *args):
        #working for db:
            if isinstance(field_val, Savable):
                field_type = field_val.annotations['data_id']
                field_val = field_val.save(True).data_id
            
            if (isinstance(field_type, type) and issubclass(field_type, Savable)\
                and not isinstance(field_val, Savable)):
                field_type = type(field_val)
                
            field_val = self.db.cast_py_db_val(field_val, field_type)

            return field_val


    def create(self, edata: EData, asynch:bool=False) -> object:
        edata.data_id = self.generate_id()
        script = self.serialize_for_create(edata)
        if asynch:
            self.cache(script)
        else:
            self.execute(script)
        return edata

    
    # def read(self, *args, **kwargs) -> Any:        
    #     # if not (args or kwargs):
    #     #     return self.read_by_fields(**kwargs)
        
    #     if kwargs:
    #         return self.read_by_kwargs(*args, **kwargs)

    #     if args:
    #         selector = args[0]
    #     else:
    #         # args.__add__(self.querying_proxy)
    #         selector = self.querying_proxy@True
    #         args = args + (selector,)

    #     if isinstance(selector, str):
    #         return self.read_by_sql(selector)

    #     if callable(selector):
    #         return self.read_by_function(selector)

    #     if isinstance(args[-1], Term):
    #         selector = args[-1]
    #         return self.read_by_term(args[0:-1], selector)
        
    #     if args:
    #         res = self.serialize_read(header=args[0], joins=[], where_clause=args[1], full_names=True)
    #         res = res.replace(r"\\","\\")
    #         return self.read_by_sql(res)

    #     raise NotImplementedError(f'couldn read by object of type {type(selector)}')

    def read(self, *args, **kwargs) -> Any:

        # if kwargs:
        #     return self.read_by_kwargs(*args, **kwargs)
        
        header = [self.querying_proxy]
        where_clause = None
        joins = []

        if len(args) >= 3:
            header = args[0]
            joins = args[1]
            where_clause = args[2]
        
        elif len(args) == 2:
            header = args[0]
            where_clause=args[1]
        
        elif len(args) == 1:
            where_clause = args[0]

        # res = self.serialize_read(header=header, joins=[], where_clause=where_clause, full_names=True)
        res = self.select(header, where_clause, joins=joins,include_data_id=True, **kwargs)
        # res = res.replace(r"\\","\\")
        return self.read_by_sql(res)

        raise NotImplementedError(f'couldn read by object of type {type(args)}')
        
    
    def delete(self, *args, **kwargs):
        return self.delete_by_id(args[0])
    
    def delete_by_id(self, data_id: str|UUID):
        delete_sql = self.serialize_for_delete(data_id)
        res = self.execute(delete_sql)
        return res


    # def read_by_fields(self):
    #     raise NotImplementedError

    # def read_by_kwargs(self, header:List[str]=[], operator:str="", **kwargs):
    #     term_header = []
    #     tp = self.querying_proxy
    #     for col_name in header:
    #         column = getattr(tp, col_name)
    #         term_header.append(column)
        
    #     if not term_header:
    #         term_header.append(tp)
    #     # from ...parser.leaf import Primitive 
    #     # term = term_header @ Primitive(True)
    #     kwargs_items = list(kwargs.items())
    #     first_item = kwargs_items[0]
    #     term = term_header @ getattr(tp, first_item[0]) == first_item[1]
    #     for (key, val) in kwargs_items[1:]:
    #         if operator == 'or':
    #             term = term | getattr(tp, key) == val
    #         else:
    #             term = term & getattr(tp, key) == val
    #     return self.read(term)
    
    # @escript
    # def read_by_kwargs(self, header:List[str]=[], operator:str="", **kwargs):
    #     tp = self.tp
    #     _header = deepcopy(header)
        
    #     if not _header:
    #         _header.append(tp)
            
    #     kwargs_items = list(kwargs.items())
    #     first_item = kwargs_items[0]
    #     term = getattr(tp, first_item[0]) == first_item[1]

    #     for (key, val) in kwargs_items[1:]:
    #         if operator == 'or':
    #             term = term or getattr(tp, key) == val
    #         else:
    #             term = term and getattr(tp, key) == val
        
    #     return self.read(_header, term)

    def read_by_sql(self, selector):
        res = self.execute(selector)
        res = self.deserialize(res)
        return res

    def read_by_function(self, func):
        selector = func(self.querying_proxy, self.resource_proxy)
        if isinstance(selector, list) or isinstance(selector, tuple):
            return self.read(*selector)
        return self.read(selector)

    def read_by_term(self, header, selector:Term):
        sql = self.parser.parse(header, selector)
        res = self.read(sql)
        return res

    def deserialize(self, rows: dict, lazy_read:bool=True):
        res = []
        if not rows:
            return res
        full_name_epure_dict = self._column_full_name_epure_dict(rows[0])
        for data_dict in rows:
            res_row = self._init_epures_row(data_dict, full_name_epure_dict, lazy_read)
            res.append(res_row)
        return res


    def _init_epures_row(self, data_dict, full_name_epure_dict, lazy_read) -> Dict[type, dict]:
        epure_attrs_dict = OrderedDict()
        for full_name, db_val in data_dict.items():
            column_tuple = full_name_epure_dict[full_name]
            epure_cls = column_tuple[0]
            column = column_tuple[1]
            if epure_cls not in epure_attrs_dict:
                epure_attrs_dict[epure_cls] = {}

            attrs = epure_attrs_dict[epure_cls]            
            # db_val = node_dict[full_name]
            if isinstance(column, TableColumn):
                field_type = column.py_type
                if isinstance(field_type, Constraint):
                    field_type = field_type.py_type
                val = self.db.cast_db_py_val(db_val, field_type)
                attrs[column.name] = val
            else:
                val = self.db.cast_db_py_val(db_val)
                attrs[column] = val

        epure_dict_items = epure_attrs_dict.items()

        if len(epure_dict_items) == 1:
            item = list(epure_dict_items)[0]
            return self._init_epure(item[0], item[1], lazy_read)

        res = []
        for epure_cls, attrs in epure_dict_items:
            epure_obj = self._init_epure(epure_cls, attrs, lazy_read)
            res.append(epure_obj)
        return res


    def _init_epure(self, epure_cls:Epure, attrs:dict, lazy_read):

        res = None
        if True in epure_cls.init_params:
            res = epure_cls(**attrs)
        else:
            arguments = {}
            for name in epure_cls.init_params:
                if isinstance(name, str) and name in attrs:
                    arguments[name] = attrs[name]
            try:
                res = epure_cls(**arguments)
            except Exception as ex:
                raise ex
        if res is None:
            raise DeserializeError
        
        for field_name, field_val in attrs.items():
            setattr(res, field_name, field_val)

        if not (hasattr(res, 'annotations') and 'data_id' in attrs):
            return res

        res.__promises_dict__ = {}

        for field_name, field_type in res.annotations.items():
            from ..edata.elist import ECollectionMetacls
            from ...resource.edata.elist import Elist, Eset


            if self.is_excluded(epure_cls, field_name, field_type):
                continue

            if field_name not in attrs:
                data_id = attrs['data_id']
                promise = FieldPromise(epure_cls.resource, data_id, field_name)
                # setattr(res, field_name, promise)
                # delattr(res, field_name)
                res.__promises_dict__[field_name] = promise

            elif attrs[field_name] == None:
                setattr(res, field_name, None)

            elif isinstance(field_type, Epure):
                data_id = attrs[field_name]
                if issubclass(field_type, Proto) and field_name == '__proto__':
                    proto = field_type.resource.read(data_id=data_id)
                    res.set_proto_fields(proto)
                elif lazy_read:
                    promise = DataPromise(field_type.resource, data_id)
                    # setattr(res, field_name, promise)
                    delattr(res, field_name)
                    res.__promises_dict__[field_name] = promise
                elif not lazy_read:
                    edata = field_type.resource.read(data_id=data_id)
                    setattr(res, field_name, edata)

            elif isinstance(field_type, ECollectionMetacls) and field_type.__origin__ == Elist:
                eset_id = attrs[field_name]

                collection_epure = field_type.get_collection_epure(parent_obj=res, field_name=field_name)
                
                if lazy_read:
                    promise = ElistPromise(collection_epure.resource, eset_id, field_type)
                    # setattr(res, field_name, promise)
                    delattr(res, field_name)
                    res.__promises_dict__[field_name] = promise
                elif not lazy_read:
                    list_values_rows = collection_epure.resource.read(eset_id=eset_id)
                    edata = field_type(list_values_rows)
                    setattr(res, field_name, edata)

            elif isinstance(field_type, ECollectionMetacls) and field_type.__origin__ == Eset:
                eset_id = attrs[field_name]

                collection_epure = field_type.get_collection_epure(parent_obj=res, field_name=field_name)
                
                if lazy_read:
                    promise = EsetPromise(collection_epure.resource, eset_id, field_type)
                    # setattr(res, field_name, promise)
                    delattr(res, field_name)
                    res.__promises_dict__[field_name] = promise
                elif not lazy_read:
                    list_values_rows = collection_epure.resource.read(eset_id=eset_id)
                    edata = field_type(list_values_rows)
                    setattr(res, field_name, edata)

        return res



    def _column_full_name_epure_dict(self, data_dict) -> Dict[str, tuple]:
        res = OrderedDict()
        for full_name, val in data_dict.items():
            dot_name = full_name.replace('___', '.')
            split = dot_name.rsplit('.', 1)
            table_name = split[0]
            field_name = split[1]            
            epure_cls = self.db.get_epure_by_table_name(table_name)
            if not epure_cls:
                res[full_name] = (object, field_name)
                continue

            epure_table = getattr(epure_cls, 'resource', None)
            if epure_table == None or not isinstance(epure_table, Table):
                raise DbError(f'epure {epure_cls} must have table as resourse')
            column = epure_table.header[field_name]
            res[full_name] = [epure_cls, column]
        return res


    def update(self, edata: Savable, asynch:bool=False) -> object:
        script = self.serialize_for_update(edata)
        if asynch:
            self.cache(script)
        else:
            self.execute(script)
        return edata


    def cache(self, script: str):        
        self.db.cache_queue.append(script)


    def _set_header(self, header):
        if header == None:
            header = TableHeader(table=self)
        
        if isinstance(header, Dict):
            header = TableHeader(columns=header, table=self)
        
        self.header = header
        self.header.resource = self
       
    def serialize_for_create(self, edata: Savable, **kwargs) -> object:
        raise NotImplementedError

    def serialize_for_read(self, edata: Savable, **kwargs) -> object:
        raise NotImplementedError

    def serialize_for_update(self, edata: Savable, **kwargs) -> object:
        raise NotImplementedError

    def serialize_for_delete(self, *args, **kwargs) -> object:
        raise NotImplementedError

    
    def serialize_header(self, db: TableStorage=None, **kwargs) -> List[Dict[str, str]]:
        res: List[Dict[str, str]] = list()

        if db == None:
            db = self.db

        if db == None:
            raise EpureError('undefined db for header serialization')
        
        header = self.header
        for column_name in header:
            column = header[column_name]
           
            serialized = {
                "column_name": column.name,
                "column_type": column.serialize_type(db) #db.get_db_type(column.column_type)
            }
            res.append(serialized)
        return res

    def _add_data_id_fields(self, header:List):
        res = []
        for item in header:
            # if isinstance(item, ColumnProxy):
            if isinstance(item, ColumnProxyBase):
                md = item.__model__
                data_id = getattr(md, 'data_id', None)
                if data_id is not None and not data_id.in_header(header + res):
                    res.append(data_id)
        
        # res = tuple(res)
        header = tuple(header)
        
        if isinstance(header[0],str):
            res = set()
            for item in header:
                sp = item.split('.')

                if len(sp) == 3:
                    data_id_column = f'{sp[0]}.{sp[1]}.data_id'
                    if data_id_column not in header:
                        res.add(data_id_column)
                    
        res = tuple(res)



        return header + res


    # def get_column_header_name(self, column_name:str):
    #     raise NotImplementedError

        # tables = []
        # columns = []
        # for item in header:
        #     if isinstance(item, Model):
        #         table_name = item.str(False, True)
        #         tables.append(table_name)
        #     if isinstance(item, ColumnProxy):
        #         column_name = item.str(False, True)
        #         columns.append(column_name)

        # res = []
        # for item in header:
        #     if isinstance(item, ColumnProxy):
        #         tp = item.__model__
        #         table_name = tp.str(False, True)
        #         if (not table_name in tables) and\
        #                 hasattr(tp, 'node_id') and\
        #                 tp.node_id.str(False, True) not in columns:
        #             res.append(tp.node_id)
        #             columns.append(tp.node_id.str(False, True))
        # return res

    def select(self, *args, joins=[], include_data_id=False, **kwargs):

        # header = get_select_header(args[0])
        body = args[1]
        
        if kwargs:
            # return get_condition_by_kwargs(header, kwargs)
            body = self.get_condition_by_kwargs(**kwargs)

    # if args:
        return self.serialize_read(header=args[0], joins=joins, where_clause=body, full_names=True, include_data_id=include_data_id)
        
    # @escript
    # def get_condition_by_kwargs(self, operator:str="", **kwargs):

    #     kwargs_items = list(kwargs.items())
    #     first_item = list(kwargs_items[0])
    #     # term = getattr(tp, first_item[0]) == first_item[1]

    #     if type(first_item[1]) in (str, UUID):
    #         first_item[1] = repr(str(first_item[1]))

    #     term = f"{prefix}.{first_item[0]} = {first_item[1]}"

    #     for (key, val) in kwargs_items[1:]:
    #             if type(val) in (str, UUID):
    #                 val = repr(str(val))

    #             if operator == 'or':
    #                 term += f" OR {prefix}.{key} = {val}"
    #             else:
    #                 term += f" AND {prefix}.{key} = {val}"

    #     return term

    @escript
    def get_condition_by_kwargs(self, operator="", **kwargs):
        md = self.md
            
        kwargs_items = list(kwargs.items())
        first_item = kwargs_items[0]
        term = getattr(md, first_item[0]) == first_item[1]

        for (key, val) in kwargs_items[1:]:
            if operator == 'or':
                term = term or getattr(md, key) == val
            else:
                term = term and getattr(md, key) == val
        
        return term

    # def get_select_header(header):
        
    #     if not isinstance(header, collections.abc.Sequence):
    #         header = tuple(header)
        
    #     return header


from __future__ import annotations
from types import NoneType, CodeType
from typing import cast, Callable, List, Any, get_type_hints, Dict, TYPE_CHECKING
from .resource.resource import Resource
from .helpers.type_helper import check_type
from .errors import EpureError, DefaultConstraintError
from .resource.db.constraint import Foreign, Default, Constraint
from .resource.edata.ecollection_metacls import ECollectionMetacls

from uuid import UUID
from .parser.inspect_parser.domain import Domain
from .parser.inspect_parser.model import Model
import textwrap
import inspect
from .parser.inspect_parser.inspect_parser import InspectParser
import types
import ast
from importlib import import_module
import os
from typing import Any, Callable, TypeVar

DecoratedCallable = TypeVar("DecoratedCallable", bound=Callable[..., Any])

# from types import FunctionType
from .resource.edata.edata import TableData
from .resource.savable import Savable
if TYPE_CHECKING:
    from .resource.db.table_storage import TableStorage
    from .resource.db.table import Table
from .resource.db.db_entity import DbEntity
import functools
from .parser.term import Term
from .resource.edata.proto import Proto
from inspect import signature



class Epure(type, Savable):
    
    EDb:TableStorage
    epures:List[Epure] = []
    # annotations:Dict[str,Any]
    resource:Savable
    prepared_resource:object
    init_params:object
    
    # def __new__(mcls, name: str, cls_bases: Tuple[type, ...], namespace: dict[str, Any]) -> Epure:
    #     return super().__new__(mcls, name, cls_bases, namespace)    
    
    def __call__(self, *args, **kwargs):
        if not self.is_saved:
            self.save_epure()
            self.is_saved = True    # __getattr__ is called here, you cannot enter __getattr__ with debuger, so put breakpoint on start of getattr func

        res = super(Epure, self).__call__(*args, **kwargs)
        return res


    def __getattr__(self, attr_name: str) -> Any:
        if not self.is_saved:
            self.is_saved = True    # __getattr__ is called here, you cannot enter __getattr__ with debuger, so put breakpoint on start of getattr func
            self.save_epure()
            return getattr(self, attr_name)
        raise AttributeError(f"'{type(self)}' object has no attribute '{attr_name}'")
        # if self.is_saved:
        #     raise AttributeError(f"'{type(self)}' object has no attribute '{attr_name}'")
        # self.is_saved = True
        # self.save_epure()        
        # return getattr(self, attr_name)

    #decorator
    # @classmethod
    # def read(cls, method):
    #     querying_proxy = None
    #     resource_proxy = None
    #     if getattr(cls, 'resource', None):
    #         querying_proxy = getattr(cls.resource, 'querying_proxy', None)
    #         resource_proxy = getattr(cls.resource, 'resource_proxy', None)

    #     @functools.wraps(method)
    #     def wrap(self, *args, **kwargs):
    #         res = None
    #         if querying_proxy and resource_proxy:
    #             res = method(self, querying_proxy, resource_proxy, *args, **kwargs)
    #         if querying_proxy:
    #             res = method(self, querying_proxy, *args, **kwargs)
    #         if resource_proxy:
    #             res = method(self, resource_proxy, *args, **kwargs)
    #         else:
    #             res = method(self, *args, **kwargs)
    #         if isinstance(res, Term):
    #             return self.resource.read(res)
    #         return res
    #     return wrap
        # def reader(self:Table, *args, **kwargs):
        #     pseudo_self = PresudoTable(self)
        #     pseudo_db = PseudoDb(self.db)
        #     script = selector(pseudo_self, pseudo_db, *args, **kwargs)
        #     return self.execute(selector)
        # setattr(self, selector.__name__, reader)

    def prepare_save(self, resource:object):
        self.__class__.epures.append(self)
        self.prepared_resource = resource


    def save_epure(self):
        self.set_resource()
        self.init_params = self.get_init_params()

    def get_init_params(self):
        res = []
        sig = signature(self.__init__)
        for key, val in sig.parameters.items():
            if val.kind == val.VAR_POSITIONAL:
                    continue
            if val.kind == val.VAR_KEYWORD:
                res.append(True)
            else:
                res.append(key)
        return res[1:]
        # sig = signature(epure_cls.__init__)
        # params = sig.parameters
        # params_vals = params.values()
        # has_kwargs = any([True for p in params_vals if p.kind == p.VAR_KEYWORD])
        # if has_kwargs:
        #     res = epure_cls(**kwargs)
        # else:
        #     arguments = {}
        #     for key, val in params:
        #         if val.kind == val.VAR_POSITIONAL:
        #             continue
        #         if key in kwargs:
        #             arguments[key] = kwargs[key]
        #             continue
        #         if val.default == val.empty:
        #             raise DeserializeError
        #     res = epure_cls(**arguments)


    def set_resource(self):
        resource = self.prepared_resource
        check_type('resource', resource, [str, Savable, NoneType])

        if type(resource) == Savable:
            resource = cast(Savable, resource)
            if not hasattr(resource, 'resource'):
                raise EpureError('For using savable as resource for epure, savable must have own resource')
            grandpa = resource.resource
            grandpa.update(self)
        else:
            table_name = self._get_table_name(self, resource)
            resource = self._create_or_update(table_name)

        self.resource = resource

        

    def _create_or_update(self, table_name:str) -> Table:
        from .resource.db.table import Table
        table:Table = self._get_table_by_cls(table_name)
        res:DbEntity
        if table_name in self.EDb:
            deleted_columns = self.EDb[table_name].header.deleted_columns
            table.header.deleted_columns = deleted_columns
            res = self.EDb.update(table)
        else:
            res = self.EDb.create(table)
        res = cast(Table, res)
        return res

    def _get_table_name(self, cls, resource: object) -> str:
        
        if not isinstance(resource, str):
            resource = cast(Savable, resource)
            return resource.full_name

        if resource:
            return resource
        
        return self.EDb._get_full_table_name(cls.__name__).full_name

        

    def _get_table_by_cls(self, table_name: str = '') -> Table:
        from .resource.db.table import Table
        table:Table
        
        full_name = self.EDb._get_full_table_name(table_name)        
        columns = {}

        for field_name, py_type in self.annotations.items():
            py_type = cast(type, py_type)

            if Table.is_excluded(self, field_name, py_type):
                continue
            py_type = self.get_py_type(field_name, py_type)
            
            columns[field_name] = py_type

        if len(columns) == 1 and list(columns.keys())[0] == 'data_id':
            TableCls = self.EDb.nosql_table_type
        else:
            TableCls = self.EDb.default_table_type
        table = TableCls(full_name.name, columns, self.EDb, full_name.namespace, parser=self.EDb.parser)
        return table

    def get_py_type(self, field_name:str, py_type:type) -> type:
        # from .resource.edata.elist import ElistMetacls
        # from .resource.edata.elist import ECollectionMetacls


        if py_type in self.epures:
            # py_type = cast(Epure, py_type)
            return self.get_py_type(field_name, py_type.annotations['data_id'])
        
        # if isinstance(py_type, ElistMetacls):
        if isinstance(py_type, ECollectionMetacls):
            return self.get_py_type(field_name, UUID)
        # if isinstance(py_type, Constraint)
            
        if isinstance(py_type, Constraint) and issubclass(py_type.__origin__, Default):
            return self.get_default_type(field_name, py_type)

        return py_type

    def get_default_type(self, field_name:str, py_type:Default) -> Any:
        default = getattr(self, field_name, None)
        if not (default or py_type.default):
            raise DefaultConstraintError(self.full_name, field_name)
        if not py_type.default:
            py_type.default = default

        if not isinstance(py_type.default, py_type.py_type):
            raise DefaultConstraintError(message=f'''field {field_name} of class {self.full_name} 
                    has wrong defaul value {py_type.default} not corresponding to {py_type.py_type} type''')

        return py_type


    def get_foreign_type(self, foreign:Epure) -> Any:
        foreign_id_type = foreign.annotations['data_id']

        foreign_table = self._get_table_name(foreign, foreign.prepared_resource)

        foreign_column = 'data_id'
        column_type = Foreign[foreign_id_type, foreign_table, foreign_column]
        return column_type


        

def epure(resource:object='', saver:type=TableData, epure_metaclass:type=Epure) -> Callable:
    check_type('resource', resource, [str, Savable, NoneType])

    def epure_creator(cls:type) -> Epure:
        epure_cls = _create_epure(cls, saver, epure_metaclass)
        epure_cls.is_saved = False      # might not change true to false in the debuger, because this is overrided setter.
        epure_cls.prepare_save(resource)

        del cls
        return epure_cls

    return epure_creator


def _create_epure(cls, saver, _Epure):
    cls_dict = dict(cls.__dict__)
    cls_dict.pop('__dict__', None)
    if issubclass(cls, Savable):
        return _Epure(cls.__name__, cls.__bases__, cls_dict)

    bases = list(cls.__bases__)
    if object in bases:
        bases.remove(object)
    bases.append(saver)
    return _Epure(cls.__name__, tuple(bases), cls_dict)


def proto(resource:object='', saver:type=Proto, epure_metaclass:type=Epure) -> Callable:
    return epure(resource, saver, epure_metaclass)

def _get_func_br_lines(func_str:str) -> Dict[int, str]:
    
    code_src_splited = func_str.splitlines(True)

    src_lines_dict = dict(enumerate(code_src_splited))

    # br_lines_list = {k for k, v in src_lines_dict.items() if v=='\n' or v.endswith('\\\n') or ("#" in v and v.strip().startswith('#'))}
    
    br_lines_dict = {}

    for key, val in src_lines_dict.items():
        if val =='\n' or ("#" in val and val.strip().startswith('#')):
            br_lines_dict[key] = "\n"
        
        elif val.endswith('\\\n'):
            br_lines_dict[key] = "\\\n"


    return br_lines_dict

def _insert_br_lines(func_str:str, br_lines_dict:Dict[int, str]) -> str:

    new_func_list = func_str.splitlines(True)

    for num, br_char in br_lines_dict.items():
        new_func_list.insert(num, br_char)

    func_str = "".join(new_func_list)

    return func_str

def _compile_func(func:Callable, new_func_w_br_lines:str, func_name:str) -> CodeType:

    file_name = func_name + ".py"

    code_block = compile(new_func_w_br_lines, file_name, 'exec')

    i = next(i for i, v in enumerate(code_block.co_consts) if isinstance(v, CodeType))

    fn_code = code_block.co_consts[i]
    n_fn_code = func.__code__

    # if 'InspectParserTestCls2' in list(fn_code.co_names): fn_code.names = tuple(list(fn_code.co_names).remove('InspectParserTestCls2', None))

    # new_co_names = tuple(filter(lambda item: item not in n_fn_code.co_freevars, fn_code.co_names))

    new_code = CodeType(
    fn_code.co_argcount,
    fn_code.co_posonlyargcount,
    fn_code.co_kwonlyargcount,
    fn_code.co_nlocals,
    fn_code.co_stacksize,
    fn_code.co_flags,
    fn_code.co_code,
    fn_code.co_consts,
    fn_code.co_names,
    # new_co_names,
    fn_code.co_varnames,
    n_fn_code.co_filename,
    fn_code.co_name,
    n_fn_code.co_firstlineno,
    fn_code.co_linetable,
    # fn_code.co_lnotab,
    n_fn_code.co_freevars,
    n_fn_code.co_cellvars)

    return new_code

def escript(func: Callable) -> Callable[[DecoratedCallable], DecoratedCallable]:

    func_source = inspect.getsource(func)

    dedent_src = textwrap.dedent(func_source)

    br_lines_dict = _get_func_br_lines(dedent_src)

    parsed_tree = InspectParser().parse(dedent_src)

    ast.fix_missing_locations(parsed_tree)

    new_func_str = ast.unparse(parsed_tree)

    new_func_w_br_lines = _insert_br_lines(new_func_str, br_lines_dict)

    func_name = parsed_tree.body[0].name

    new_func_code = _compile_func(func, new_func_w_br_lines, func_name)

    # global new_dict, newfunc

    func.__code__ = new_func_code

    new_dict = func.__globals__
    # newfunc = types.FunctionType(new_func_code, new_dict, name=func.__name__, argdefs=func.__defaults__, closure=func.__closure__)

    def inner(self, *args, **kwargs) -> Any:
        
        # new_dict = func.__globals__
        if self.__class__.__name__ in new_func_code.co_freevars and self.__class__.__name__ not in new_dict:
            new_dict.update({f"{self.__class__.__name__}" : self.__class__})
            nonlocal func
            func = types.FunctionType(new_func_code, new_dict, name=func.__name__, argdefs=func.__defaults__, closure=func.__closure__)

        #temporary holders for tp, db proberties
        # tp_err_prop = getattr(self, "tp")
        # dbp_err_prop = getattr(self, "tp")
        # from .resource.edata.elist import ECollectionMetacls
        # from .resource.edata.elist import Elist, Eset

        if isinstance(type(self), Epure) or isinstance(self, Epure):
            db = self.resource.db
            full_name = self.resource.full_name
        elif isinstance(type(self), ECollectionMetacls):
            db = self.collection_epure.resource.db
            full_name = self.collection_epure.resource.full_name
        else:
            db = self.db
            full_name = self.full_name
        
        self.dom = Domain(db)
        self.md = getattr(self.dom, full_name)
        self.get_model = get_model_from_class

        res = func(self, *args, **kwargs)

        delattr(self, "dom")
        delattr(self, "md")
        delattr(self, "get_model")
        
        return res

    return inner

def get_model_from_class(cls:Epure) -> Model:
    if cls not in Epure.epures:
        raise EpureError(f"Model cannot be created from class {cls.__name__} that is not decorated by @epure")
    return Model(cls.resource.db, cls.resource)


from .resource import Resource

class DataPromise:
    resource:Resource
    data_id:object
    def __init__(self, resource, data_id) -> None:
        self.resource = resource
        self.data_id = data_id

    def get(self):
        res = self.resource.read(data_id=self.data_id)[0]
        return res

class ElistPromise(DataPromise):
    py_type:type

    def __init__(self, resource, data_id, py_type) -> None:
        self.py_type = py_type
        return super().__init__(resource, data_id)

    def get(self):
        list_values = self.resource.read(eset_id=self.data_id)
        data = self.py_type(list_values)
        # node = self.py_type[self.py_type.collection_epure](list_values)
        return data
    
class EsetPromise(DataPromise):
    py_type:type

    def __init__(self, resource, data_id, py_type) -> None:
        self.py_type = py_type
        return super().__init__(resource, data_id)

    def get(self):
        # list_values = self.resource.read(eset_id=self.data_id)
        data = self.py_type([], self.resource)
        data.data_id = self.data_id
        # node = self.py_type[self.py_type.collection_epure](list_values)
        return data

class FieldPromise(DataPromise):
    field_name:str
    def __init__(self, resource, data_id, field_name) -> None:
        self.field_name = field_name
        return super().__init__(resource, data_id)

    def get(self):
        res = self.resource.read([self.field_name], data_id=self.data_id)
        return getattr(res, self.field_name)


from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...parser.inspect_parser.model import Model
    # from ...resource.db.table import Table

class JoinMethods:
    
    def left_join(self, joined_model:Model, on_clause:str, alias:str=""):
        return self._join(joined_model, on_clause, "LEFT", alias)

    def right_join(self, joined_model:Model, on_clause:str, alias:str=""):
        return self._join(joined_model, on_clause, "RIGHT", alias)

    def join(self, joined_model:Model, on_clause:str, alias:str=""):
        return self._join(joined_model, on_clause, "", alias)

    def full_join(self, joined_model:Model, on_clause:str, alias:str=""):
        return self._join(joined_model, on_clause, "OUTER", alias)

class Join:
    model:Model
    on_clause:str
    join_type:str
    alias:str

    def __init__(self, model:Model, on_clause:str, join_type:str="LEFT", alias:str="") -> None:
        self.model = model
        self.on_clause = on_clause
        self.join_type = join_type
        self.alias = alias
        
class JoinResource(JoinMethods):
    model:Model
    joins:list[Join]

    def __init__(self, model:Model) -> None:
        self.joins = []
        self.model = model

    def _join(self, joined_model:Model, on_clause:str, join_type:str="", alias:str="") -> None:
        join = Join(joined_model, on_clause, join_type, alias)
        res = JoinResource(self.model)
        res.joins = self.joins.copy()
        res.joins.append(join)
        return res

    def read(self, *args, **kwargs):

        # header = []
        where_clause = ""

        if len(args) >= 2:
            header = args[0]
            where_clause = args[1]

        elif len(args) == 1:
            where_clause = args[0]

        if len(args) <= 1:
            header = [self.model]
            header = header + [join.model for join in self.joins]

        return self.model.__table__.read(header, self.joins, where_clause, **kwargs)


from __future__ import annotations
from .term import Term
from typing import Any
from .column_proxy import ColumnProxy
from ..proxy_base_cls import ModelBase
from inflection import underscore
from ...resource.join_resource.join_resource import JoinResource, JoinMethods

class Model(Term, ModelBase, JoinMethods):

    def __init__(self, db, table):
        self.__db__ = db
        self.__table__ = table
        self.__resource__ = table
        self.__qp_name__ = self.serialize(parentheses=False, full_names=True)
        super().__init__()

    def __getattr__(self, attr_name: str) -> ColumnProxy:
        # if self.__is_copy__:
            # raise AttributeError
        
        # case_name = underscore(attr_name)
        case_name = attr_name.lower()
        
        if case_name not in self.__table__.header:
            raise AttributeError(f'column {attr_name} not in header of table {self.__table__.full_name}')
        column = self.__table__.header[case_name]
        res = ColumnProxy(self.__db__, self.__table__, column, self)
        return res

    # def __getitem__(self, *args) -> Any:
    #     res = TermHeader(list(*args))
    #     return res

    def serialize(self, parentheses=True, full_names=True, for_header=False) -> str:
        res = ''
        # if not full_names:
        #     res = self.__table__.name      
        # el
        if for_header:
            header = self.__table__.header
            for col in header:
                res += header.serialize_read_column(header[col], full_names) + ', '
            res = res[0:-2]
        else:
            res = f'{self.__table__.full_name}'

        # if not for_header and parentheses:
        #     res = self.append_parentheses(res)

        return res
    
    
    def select(self, *args, joins=[], include_data_id=False, **kwargs):
        if type(args[0]) not in (tuple,list,set):
            raise ValueError(f"first arg must be tuple, list or set, not {args[0]}")
        return "(" + self.__table__.select(*args, joins=joins, include_data_id=include_data_id, **kwargs) + ")"
    
    def _join(self, joined_model:Model, on_clause:str, join_type:str="", alias:str="") -> JoinResource:
        join_resource =  JoinResource(self)
        new_join_resource = join_resource._join(joined_model, on_clause, join_type, alias)
        return new_join_resource


from _ast import Assign, Attribute, Call, Constant, FunctionDef, In, Not, UnaryOp
import ast
from types import CodeType
from typing import Any, Dict
# import inspect
# import textwrap
from .term import Term
from .domain import Domain
from .model import Model
from .column_proxy import ColumnProxy

# class AstParser(ast.NodeTransformer):
class InspectParser(ast.NodeTransformer):

    astTypesDict:Dict[str, Any]
    column_proxy_cls=ColumnProxy
    model_cls=Model
    ast_type_method_name_dict:Dict[ast.cmpop, str] = {
        ast.Eq: "__eq__",
        ast.NotEq: "__ne__",
        ast.Lt: "__lt__",
        ast.LtE: "__le__",
        ast.Gt: "__gt__",
        ast.GtE: "__ge__",
        ast.In: "_in",
        ast.NotIn: "not_in",
        ast.Or: "_or",
        ast.And: "_and",
        ast.Is: "_is",
        ast.IsNot: "is_not"
    }

    def __init__(self) -> None:
        self.astTypesDict = {}
        super().__init__()

    def parse(self, func) -> CodeType:
        # func_source = inspect.getsource(func)
        # dedent_src = textwrap.dedent(func_source)
        func_tree = ast.parse(func)
        func_args = func_tree.body[0].args.args
        self.first_arg_name = func_args[0].arg

        attr_tp = ast.Attribute()
        attr_dbp = ast.Attribute()
        attr_tp.type = Model
        attr_dbp.type = Domain

        self.astTypesDict[f'{self.first_arg_name}.md'] = attr_tp
        # self.astTypesDict[f'{self.first_arg_name}.querying_proxy'] = attr_tp
        self.astTypesDict[f'{self.first_arg_name}.dom'] = attr_dbp

        changed_tree = self.visit(func_tree)
        fixed_tree = ast.fix_missing_locations(changed_tree)
        return fixed_tree
    
    # def visit_FunctionDef(self, node: FunctionDef) -> Any:
        
    #     self.generic_visit(node)

    #     if node.decorator_list:
    #         i = next(i for i, v in enumerate(node.decorator_list) if v.id == "escript")
    #         node.decorator_list.pop(i)

    #     return node

    # def is_bool_operand(self, operand_val, operand_val_str):
    #     res = operand_val_str in self.astTypesDict.keys()\
    #         or isinstance(operand_val, ast.Compare)\
    #         or isinstance(operand_val, ast.In)\
    #         or isinstance(operand_val, ast.NotIn)\
    #         or isinstance(operand_val, ast.BoolOp)
        
    #     return res
    
    def visit_BoolOp(self, node: ast.BoolOp) -> Any: # and or

        self.generic_visit(node)
        
        res_node_str = ""
        left_val = node.values[0]

        for index in range(1, len(node.values)):
            compare_targ = node.values[index]

            left_val_str = ast.unparse(left_val)
            compare_targ_str = ast.unparse(compare_targ)

            left_val_in_keys = left_val_str in self.astTypesDict.keys()
            comp_targ_in_keys = compare_targ_str in self.astTypesDict.keys()

            if left_val_in_keys or comp_targ_in_keys:
                op_str = self.ast_type_method_name_dict[type(node.op)]
                new_node_str = f"{self.first_arg_name}.md.{op_str}({left_val_str}, {compare_targ_str})"
                new_node = ast.parse(new_node_str).body[0].value
                new_node.type = Term
                self.astTypesDict[new_node_str] = new_node

                # new_node_str = self.switch_operator(node.op, left_val_str, compare_targ_str)
                # left_val = node
            else:
                op_str = "and" if type(node.op) == ast.And else "or"
                new_node_str = f"{left_val_str} {op_str} {compare_targ_str}"

            res_node_str = new_node_str
            left_val = ast.parse(res_node_str).body[0].value

        node = ast.parse(res_node_str).body[0].value
        
        return node
    
    def visit_UnaryOp(self, node: UnaryOp) -> Any:

        self.generic_visit(node)

        node_str = ast.unparse(node.operand)

        if type(node.op) == Not and node_str in self.astTypesDict:
            new_node_str = f"{self.first_arg_name}.md._not({node_str})"
            node = ast.parse(new_node_str).body[0].value
            node.type = Term
            self.astTypesDict[new_node_str] = node
        
        return node
    
    def visit_Compare(self, node: ast.Compare) -> Any:

        self.generic_visit(node)

        # self.generic_visit(node)
        if type(node.ops[0]) in (ast.In, ast.NotIn):
            node = self.handle_In_and_Not_In(node)
        else:
            node = self.handle_Compare_Op(node)

        return node
    
    def visit_Assign(self, node: Assign) -> Any:

        self.generic_visit(node)

        # if type(node.body[0].value) in [i.type for i in self.astTypesDict.values()]:
        #     node.body[0]

        right_val = ast.unparse(node.value)
        assign_target = ast.unparse(node.targets)

        if assign_target not in self.astTypesDict and\
        right_val in self.astTypesDict.keys() and\
        issubclass(self.astTypesDict[right_val].type, Term):
            self.astTypesDict[assign_target] = self.astTypesDict[right_val]

        if assign_target in self.astTypesDict.keys() and\
        right_val not in self.astTypesDict.keys() and\
        issubclass(self.astTypesDict[assign_target].type, Term):
            # self.astTypesDict.pop(assign_target)
            self.astTypesDict = dict(filter(lambda x: assign_target not in x[0].split('.'), self.astTypesDict.items()))

        # if hasattr(node.value, "value") and node.value.value in (True, False):
        #     node.value.type = Term
        #     self.astTypesDict[assign_target] = node.value

        return node
    
    def visit_Attribute(self, node: Attribute) -> Any:

        self.generic_visit(node)

        left_val = ast.unparse(node.value)
        
        node = self.handle_getattr(node, left_val)
        
        return node
    
    def visit_Call(self, node: Call) -> Any:

        self.generic_visit(node)
        caller = None

        if hasattr(node.func, "value"):
            caller = ast.unparse(node.func.value)

        if hasattr(node.func,"id") and node.func.id == "getattr":
            left_val = ast.unparse(node.args[0])
            node = self.handle_getattr(node, left_val)

        elif caller and caller in self.astTypesDict and\
            hasattr(node.func,"attr") and node.func.attr == "select":
            node.type = Term
            self.astTypesDict[f'{ast.unparse(node)}'] = node

        elif caller and caller in self.astTypesDict and\
            hasattr(node.func,"attr") and node.func.attr == "like":
            node.type = Term
            self.astTypesDict[f'{ast.unparse(node)}'] = node

        elif caller and caller in self.astTypesDict and\
            hasattr(node.func,"attr") and node.func.attr == "model":
            node.type = Model
            self.astTypesDict[f'{ast.unparse(node)}'] = node

        return node
    
    # def visit_Constant(self, node: Constant) -> Any:
    #     if node.value in (True, False) and\
    #     str(node.value) not in self.astTypesDict.keys():
    #         node.type = Term
    #         self.astTypesDict[f'{ast.unparse(node)}'] = node
    #     return node
    
    def handle_getattr(self, node, left_val):

        self.generic_visit(node)

        node_str = ast.unparse(node)

        if node_str in self.astTypesDict.keys():
            return node

        if left_val in self.astTypesDict.keys() and\
        self.astTypesDict[left_val].type == Domain:
            node.type = Model
            self.astTypesDict[f'{node_str}'] = node

        if left_val in self.astTypesDict.keys() and\
        self.astTypesDict[left_val].type == Model:
            node.type = ColumnProxy
            self.astTypesDict[f'{node_str}'] = node

        return node

    def handle_In_and_Not_In(self, node: In) -> Any:
        self.generic_visit(node)

        left_val = ast.unparse(node.left)
        compare_targ = ast.unparse(node.comparators)

        left_val_in_keys = left_val in self.astTypesDict.keys()
        comp_targ_in_keys = compare_targ in self.astTypesDict.keys()

        if not left_val_in_keys:
            return node
        
        if left_val_in_keys: 
            left_val_type = self.astTypesDict[left_val].type

        if comp_targ_in_keys:
            comp_targ_type = self.astTypesDict[compare_targ].type

        if left_val_in_keys and left_val_type in (Model, Domain):
            raise TypeError(f"'{left_val}' is of type '{left_val_type}' and not of type '{ColumnProxy}', so it cannot be present in '{compare_targ}'")
        
        # if not (comp_targ_in_keys and isinstance(self.astTypesDict[compare_targ], ast.Call) and node.comparators[0].func.attr == "select") and (type(node.comparators[0]) not in (ast.Name, ast.List, ast.Tuple, ast.Set)):
            # comp_targ_type = type(node.comparators[0])
            # raise TypeError(f"'{compare_targ}' is of type '{comp_targ_type}' and not of type List, Tuple, Set or select method, so it cannot be right operand for SQL 'IN' operator")
        
        if left_val_in_keys and issubclass(left_val_type, ColumnProxy):
            op_method = self.ast_type_method_name_dict[type(node.ops[0])]
            new_node_str = f"{left_val}.{op_method}({compare_targ})"
            node = ast.parse(new_node_str).body[0].value
            node.type = Term
            self.astTypesDict[new_node_str] = node

        return node
    
    # def handle_Eq_and_Like(self, node: Eq) -> Any:
    def handle_Compare_Op(self, node: Any) -> Any: # >=, <=, == 

        self.generic_visit(node)

        new_node_str = ""
        
        left_val = ast.unparse(node.left)
        compare_targ = ast.unparse(node.comparators)

        left_val_in_keys = left_val in self.astTypesDict.keys()
        comp_targ_in_keys = compare_targ in self.astTypesDict.keys()

        if left_val_in_keys: 
            left_val_type = self.astTypesDict[left_val].type

        if comp_targ_in_keys:
            comp_targ_type = self.astTypesDict[compare_targ].type

        if not ((left_val_in_keys and issubclass(left_val_type, ColumnProxy)) or\
            (comp_targ_in_keys and issubclass(comp_targ_type, ColumnProxy))):
            return node

        if left_val_in_keys and left_val_type in (Model, Domain):
            raise TypeError(f"'{left_val}' is of type '{left_val_type}' and not of type '{ColumnProxy}', so it cannot be compared to '{compare_targ}'")

        elif comp_targ_in_keys and comp_targ_type in (Model, Domain):
            raise TypeError(f"'{compare_targ}' is of type '{comp_targ_type}' and not of type '{ColumnProxy}', so it cannot be compared to '{left_val}'")

        # if left_val_in_keys and ((compare_targ.count('%') > compare_targ.count('\%'))\
        # or (compare_targ.count('_') > compare_targ.count('\_'))):

        # if left_val_in_keys and (compare_targ.count('%') > compare_targ.count('\%'))\
        #     and issubclass(left_val_type, ColumnProxy):
        #     # compare_targ = ast.unparse(self.handle_Like(node.comparators))
        #     new_node_str = f"{left_val}._like({compare_targ})"

        # elif comp_targ_in_keys and (left_val.count('%') > left_val.count('\%'))\
        #     and issubclass(comp_targ_type, ColumnProxy):
        #     # compare_targ = ast.unparse(self.handle_Like(node.comparators))
        #     new_node_str = f"{compare_targ}._like({left_val})"
        op_method = self.ast_type_method_name_dict[type(node.ops[0])]

        if left_val_in_keys and issubclass(left_val_type, ColumnProxy):
            new_node_str = f"{left_val}.{op_method}({compare_targ})"
            # new_node_str = f"{left_val} == {compare_targ}"

        elif comp_targ_in_keys and issubclass(comp_targ_type, ColumnProxy):
            new_node_str = f"{compare_targ}.{op_method}({left_val})"
            # new_node_str = f"{compare_targ} == {left_val}"

        if new_node_str and (left_val_in_keys or comp_targ_in_keys):
            new_node = ast.parse(new_node_str)
            node = new_node.body[0].value
            node.type = Term
            self.astTypesDict[new_node_str] = node
        
        return node

    # def visit_Str(self, node: Str) -> Any:
        
        # str_unparsed = ast.unparse(node)

        # if "\\" in str_unparsed:
        #     node.value = node.value.replace(r"\\","\\")

        # return node


from .term import Term
from ..proxy_base_cls import ColumnProxyBase
from typing import Union

class ColumnProxy(Term, ColumnProxyBase):
    def __init__(self, db, table, column, model=None):
        self.__db__ = db
        self.__table__ = table
        self.__column__ = column
        self.__qp_name__ = self.serialize(parentheses=False, full_names=True)

        if model is None:
            from .model import Model
            model = Model(db, table)
        self.__model__ = model
        super().__init__()

    def serialize(self, parentheses=True, full_names=True, for_header=False) -> str:
        if for_header:
            res = self.__table__.header.serialize_read_column(self.__column__, full_names)
        elif not full_names:
            res = self.__column__.name
        else:
            table = self.__table__.full_name
            res = f'{table}.{self.__column__.name}'

        # if not for_header and parentheses:
        #     res = self.append_parentheses(res)

        return res
    
    def in_header(self, header:Union[list,tuple]) -> bool:
        table_name = self.__table__.full_name
        from .model import Model
        for qp in header:
            # check_type('qp', qp, [Model, ColumnProxy])
            if isinstance(qp, ColumnProxy):
                if self.__qp_name__ == qp.__qp_name__:
                    return True
            if isinstance(qp, Model):
                if qp.__qp_name__ == table_name:
                    return True
        return False


from .term import Term
from .model import Model
from ...errors import DbError

class Domain(Term):
    def __init__(self, db, default_namespace:str=None):
        self.__db__ = db
        self.default_namespace = default_namespace
        super().__init__()

    # def __getitem__(self, key:str):
    #     if key not in self.__db__:
    #         raise DbError(f'table {key} not in db {self.__db__.full_name}')
    #     res = Model(self.__db__, self.__db__[key])
    #     return res
    
    def __getattr__(self, key:str) -> Model:
        if self.default_namespace:
            key = f"{self.default_namespace}.{key}"

        if key not in self.__db__ and key not in self.__db__.namespaces:
            raise DbError(f'table {key} not in db {self.__db__.full_name}')
        elif key in self.__db__.namespaces:
            return Domain(self.__db__, default_namespace=key)
        res = Model(self.__db__, self.__db__[key])
        return res

    def __iter__(self):
        return self.__db__.__iter__()

    def __len__(self):
        return self.__db__.__len__()
    
    # def select(self, args)

    # def model(self, cls) -> Model:
    #     # if cls in Epure.epures:
    #     return Model(cls.resource.db, cls.resource)


from .epure import epure as epure
from .epure import Epure as Epure
# from .resource.gres.gres_db import GresDb as GresDb
# from .resource.db.constraint import Constraint, Default, Foreign, Prim, NotNull, Uniq, Check
# from .resource.db.constraint import Constraint as Constraint
from .resource.edata.elist import Elist as Elist
from .resource.edata.elist import Eset as Eset
from .epure import escript as escript
from .parser.inspect_parser.model import Model as Model 
from .parser.inspect_parser.column_proxy import ColumnProxy as ColumnProxy


from __future__ import annotations
from collections.abc import Iterable, Iterator
from .edata import TableData
from typing import List, Any, Type, Generic, Set, Dict
from uuid import UUID, uuid4
from ...epure import Epure, epure, escript
from types import NoneType
from .ecollection_metacls import ECollectionMetacls
from ..db.table import Table
from ...resource.savable import Savable
from typing import get_origin

class Elist(TableData, List, metaclass=ECollectionMetacls):
# class Elist(TableNode, List):
    entries:List
    deleted_entries:List
    py_type:type = NoneType
    collection_epure:Epure = None
    data_id:UUID

    def __init__(self, _list:List=None) -> None:
        self.entries = []
        self.deleted_entries = []

        if _list == None:
            _list = []

        # if len(_list) and isinstance(_list[0], self.collection_epure):
        if len(_list) and hasattr(_list[0], "resource") and\
        _list[0].resource.full_name == self.collection_epure.resource.full_name:
            self.data_id = _list[0].eset_id
            _list.sort(key= lambda x: x.value_order)
            self.entries = _list
        else:
            for val in _list:
                self.append(val)
    
    @classmethod
    def get_collection_epure(cls, table=None, parent_obj=None, field_name=None) -> Epure:  
        return cls.collection_epure

    def save(self, asynch:bool=False) -> Elist:
            
        if not hasattr(self,"data_id") or not self.data_id:
            self.data_id = uuid4()

        val_len = self.entries.__len__()
        for i in range(val_len):
            item = self.entries[i]

            if get_origin(self.py_type) != None: # check if class is Generic
                self.py_type = get_origin(self.py_type)

            if not isinstance(item.value, self.py_type):
                raise TypeError(f"value '{item.value}' of type '{type(item.value)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
            
            item.eset_id = self.data_id
            item.value_order = i

            # if hasattr(val, "__deleted__") and val.__deleted__:
            #     self.resource.delete(val)
            if i != val_len-1:
                item.save(asynch=True)
            else:
                item.save(asynch=asynch)

        for i in range(self.deleted_entries.__len__()):
            item = self.deleted_entries[i]
            self.collection_epure.resource.delete(item.data_id)
        
        self.deleted_entries = []

        return self

    def __setitem__(self, index, item) -> None:
        if not isinstance(item, self.py_type):
                raise TypeError(f"value '{item}' of type '{type(item)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
        self.entries[index].value = item
        # res = self.collection_epure()
        # res.value_order = index
        # res.value = item
        # self.entries.__setitem__(index, res)

    # def __delitem__(self, key)

    def insert(self, index, item) -> None:
        if not isinstance(item, self.py_type):
                raise TypeError(f"value '{item}' of type '{type(item)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
        res = self.collection_epure()
        res.value_order = index
        res.value = item
        self.entries.insert(index, res)

    def append(self, item) -> None:
        # if not isinstance(item, self.py_type):
        if hasattr(item, "resource") and item.resource.full_name != self.py_type.resource.full_name:
                raise TypeError(f"value '{item}' of type '{type(item)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
        res = self.collection_epure()
        res.value_order = len(self.entries)
        res.value = item
        self.entries.append(res)

    def count(self) -> int:
        return self.entries.count()
    
    def copy(self) -> list:
        return self.entries.copy()
    
    def clear(self) -> None:
        entries = self.entries.copy()
        self.entries.clear()
        self.deleted_entries.extend(entries)

    def index(self, value:Any, *args) -> int:
        return self.entries.index(value, *args)

    def extend(self, __iterable: Iterable) -> None:
        item = next(iter(__iterable), False)
        if not ((hasattr(__iterable, "py_type") and item.py_type == self.py_type)\
                or type(item) == self.py_type):
            raise TypeError(f"value '{item}' of type '{type(item)}' is not same "
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
        self.entries.extend(__iterable)
    
    def pop(self, index = -1) -> Any:
        res = self.entries.pop(index)
        self.deleted_entries.append(res)
        return res
    
    def remove(self, __value: Any) -> None:
        for x in self.entries:
            if x.value == __value:
                __value = x
                break
        self.entries.remove(__value)
        self.deleted_entries.append(__value)
    
    def reverse(self) -> None:
        self.entries.reverse()
    
    def sort(self, *, key=None, reverse=False) -> None:
        self.entries.sort(key=key, reverse=reverse)

    def __getitem__(self: Type, _param: Any):
        return self.entries[_param].value
    
    def __repr__(self):
        return str(self.entries)
    
    # def __eq__(self, __value: Elist|object) -> bool:
    #     if issubclass(__value,Elist):
    #         return self.entries.__eq__(__value.entries)
    #     return self.entries.__eq__(__value)

    @escript
    def load(self, *args, **kwargs) -> None:
        """
        Loads all items of Elist contents like Epure, Elist, Eset values from DataPromises
        """
        if not isinstance(self.py_type, Epure):
            return

        data_id_dict:dict = {}
        
        for val in self.entries:
            if hasattr(val, '__promises_dict__') and 'value' in val.__promises_dict__:
                data_id_dict[val.__promises_dict__['value'].data_id] = val

        if len(data_id_dict) == 0:
            return

        # res = self.py_type.resource.read(lambda tp, dp: tp.node_id >= list(node_id_dict.keys()))
        # res = self.py_type.resource.read(self.tp.node_id in list(node_id_dict.keys()))
        py_type_tp = getattr(self.dom, self.py_type.resource.full_name)
        res = self.py_type.resource.read(py_type_tp.data_id in list(data_id_dict.keys()))


        for val in res:
            data_id_dict[val.data_id].value = val

    def __len__(self) -> int:
        return self.entries.__len__()
    
    @property
    def ids(self) -> List[str]:
        """
        Returns data_id's from all items of Elist
        """

        res = []
        for item in self.entries:
            # if hasattr(item, "__promises_dict__") and\
            # "value" in item.__promises_dict__:
            #     id = item.__promises_dict__['value'].node_id
            # else:
            id = item.data_id

            res.append(str(id))

        return res
    
    def __iter__(self) -> Iterator:
        return iter([i.value for i in self.entries])

class Eset(set, TableData, metaclass=ECollectionMetacls):
    # entries:Set
    deleted_entries:Set
    py_type:type = NoneType
    collection_epure:Epure = None
    data_id:UUID

    def __init__(self, _set:Set=None, collection_resourse=None, **kwargs:dict[UUID:UUID]) -> None:
        super(set, self).__init__()
        if collection_resourse != None:
            self.collection_epure = self.get_collection_epure(collection_resourse)

        if _set == None:
            _set = []

        self.deleted_entries = []
        # if len(_set) and isinstance(_set[0], self.collection_epure):
        if len(_set) and hasattr(_set[0], "resource") and\
        _set[0].resource.full_name == self.collection_epure.resource.full_name:
                # self.data_id = _set[0].eset_id
                self.update(_set, ids_dict=kwargs)
        else:
            for index, item in enumerate(_set):
                if kwargs and kwargs["ids_dict"]:
                    self.add(item, ids_dict=list(kwargs["ids_dict"].items())[index])
                else:
                    self.add(item)


    def add(self, item: Any, **kwargs) -> None:

        if not isinstance(item, (self.py_type, Savable)):
                raise TypeError(f"value '{item}' of type '{type(item)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
        
        elif isinstance(self.py_type, Savable) and isinstance(item, Savable)\
        and not self.py_type.resource.full_name == item.resource.full_name:
            raise TypeError(f"value '{item}' of type '{type(item)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
        
        # super(Eset, self).add(item)
        # self.add(item)
        res = self.collection_epure()
        res.value = item

        if isinstance(self.py_type, Savable):
            self_iter_obj = super(self.__class__, self).__iter__()
            # for val in self:
            for val in self_iter_obj:
                if item is val:
                    raise ValueError(f"value {item} is already present in Eset")
                
        if kwargs and kwargs['ids_dict']:
            res.data_id = kwargs['ids_dict'][0]
            res.eset_id = kwargs['ids_dict'][1]
            
        super(self.__class__, self).add(res)

    def save(self, asynch:bool=False) -> Elist:
            
        if not hasattr(self, "data_id") or not self.data_id:
            self.data_id = uuid4()

        val_len = self.__len__()
        self_iter_obj = super(self.__class__, self).__iter__()
        # self_list = list(self)
        self_list = list(self_iter_obj)
        for i in range(val_len):
            item = self_list[i]
            if not isinstance(item.value, self.py_type):
                raise TypeError(f"value '{item.value}' of type '{type(item.value)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
            
            item.eset_id = self.data_id

            # if hasattr(val, "__deleted__") and val.__deleted__:
            #     self.resource.delete(val)
            if i != val_len-1:
                item.save(asynch=True)
            else:
                item.save(asynch=asynch)

        for i in range(self.deleted_entries.__len__()):
            item = self.deleted_entries[i]
            if hasattr(item, "data_id") and item.data_id:
                self.collection_epure.resource.delete(item.data_id)
        
        self.deleted_entries = []

        return self
    
    def remove(self, ___element: Any) -> None:
        self_iter_obj = super(self.__class__, self).__iter__()
        # for x in self:
        for x in self_iter_obj:
            if x.value == ___element:
                ___element = x
                break
        self.deleted_entries.append(___element)
        return super(self.__class__, self).remove(___element)
    
    def discard(self, el: Any) -> None:
        self_iter_obj = super(self.__class__, self).__iter__()
        for x in self_iter_obj:
        # for x in self:
            if x.value == el:
                el = x
                break
        self.deleted_entries.append(el)
        return super(self.__class__, self).discard(el)
    
    def clear(self) -> None:
        entries = self.copy()
        self.deleted_entries.extend(entries)
        return super(self.__class__, self).clear()
    
    @property
    def ids(self) -> List[str]:
        """
        Returns data_id's from all items of Elist
        """
        # if not self.is_saved:
        #     raise Exception("Your Eset is not saved, try .save()")
        self_iter_obj = super(self.__class__, self).__iter__()
        res = []
        # for item in list(self):
        for item in self_iter_obj:
            # if hasattr(item, "__promises_dict__") and\
            # "value" in item.__promises_dict__:
            #     id = item.__promises_dict__['value'].node_id
            # else:
            id = item.data_id

            res.append(str(id))

        return res
    
    def __contains__(self, __o: object) -> bool:
        self_iter_obj = super(self.__class__, self).__iter__()
        # for x in self:
        for x in self_iter_obj:
            if x.value == __o:
                return True

        # return super().__contains__(__o)


    @classmethod
    def get_collection_epure(cls, table=None, parent_obj=None, field_name=None):        
        from .edata import EsetTableData

        if table != None:
            name = table.full_name
        else:
            from ...named import SnakeCaseNamed
            name = f'{parent_obj.__class__.__name__}__{field_name}'
            name = SnakeCaseNamed(name).full_name


        if name in Epure.EDb:
            res = Epure.EDb.get_epure_by_table_name(name)
            return res
        
        # if name in ECollectionMetacls.ecollections:
        #     res = ECollectionMetacls.ecollections[name]
        #     cls.collection_epure = res
        #     return res
    
        obj = type(name, (object,), {})
        obj.__annotations__ = {"eset_id":UUID, "value":cls.py_type}
        res = epure(resource=f'ecollections.{name}', saver=EsetTableData)(obj)
        cls.collection_epure = res
        # ECollectionMetacls.ecollections[name] = res

        return res
    

    def _redefine_collection_epure(self, collection_epure:Epure) -> None:
        vals_list = []
        ids_dict = {}
        self_iter_obj = super(self.__class__, self).__iter__()
        # for item in self:
        for item in self_iter_obj:
            vals_list.append(item.value)
            if hasattr(item, "eset_id") and hasattr(item, "data_id"):
                ids_dict[item.data_id] = item.eset_id
            
        self.clear()
        self.collection_epure = collection_epure

        # for item in vals_list:
        #     self.add(item)

        self.__init__(vals_list, ids_dict=ids_dict)
    
    def load(self) -> None:
        """
        Load Eset instance's items
        """
        list_values = self.collection_epure.resource.read(eset_id=self.data_id)
        self.__init__(list_values)

        if not isinstance(self.py_type, Epure):
            return

        data_id_dict:dict = {}
        
        for val in self:
            if hasattr(val, '__promises_dict__') and 'value' in val.__promises_dict__:
                data_id_dict[val.__promises_dict__['value'].data_id] = val

        if len(data_id_dict) == 0:
            return

        # res = self.py_type.resource.read(lambda tp, dp: tp.node_id >= list(node_id_dict.keys()))
        # res = self.py_type.resource.read(self.tp.node_id in list(node_id_dict.keys()))
        py_type_tp = getattr(self.dom, self.py_type.resource.full_name)
        res = self.py_type.resource.read(py_type_tp.data_id in list(data_id_dict.keys()))


        for val in res:
            data_id_dict[val.data_id].value = val

    def __iter__(self) -> Iterator:
        it_obj = super(self.__class__, self).__iter__()
        return iter([item.value for item in it_obj])
    
    def update(self, *s: Iterable, **kwargs:[UUID, UUID]) -> None:

        fin_res = []
        if kwargs["ids_dict"]:
            for ind, _iterable in enumerate(s):
                # if not isinstance(list(item)[0], self.collection_epure):
                    # raise TypeError("Eset cannot be updated with Eset of different type")
                res = []
                for item, ids in zip(list(_iterable)[ind], kwargs["ids_dict"].items()):
                    item.data_id = ids[0]
                    item.eset_id = ids[1]
                    res.append(item)
                
                fin_res.append(res)

        super(self.__class__, self).update(*s)


from __future__ import annotations
from uuid import UUID
from ..savable import Savable
from ..resource import Resource
from ...errors import ResourceException
from typing import Any, Dict, Callable
import jsonpickle
from ..db.constraint import Constraint
from ...errors import EpureError
from .ecollection_metacls import ECollectionMetacls
from ..data_promise import DataPromise, ElistPromise, EsetPromise
from ...helpers.type_helper import is_uuid
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...epure import Epure

from typing import get_origin

class EData(Savable):

    data_id:object
    __exclude__:list
    __promises_dict__:Dict[str, Any]

    def __init__(self, data_id:object=None, resource:Resource=None) -> None:
        if data_id != None:
            self.data_id = data_id
        super().__init__(resource)

    # def __getattribute__(self, name: str) -> Any:
    #     from ..node_promise import NodePromise
    #     val = super(EData, self).__getattribute__(name)
    #     if isinstance(val, NodePromise):
    #         setattr(self, name, val.get())
    #     return super().__getattribute__(name)
    
    def __getattr__(self, name:str):
        if '__promises_dict__' in self.__dict__ and name in self.__promises_dict__:
            res = self.__promises_dict__[name].get()
            setattr(self, name, res)
            self.__promises_dict__.pop(name)
            return res
        else:
            raise AttributeError(f"'{type(self)}' object has no attribute '{name}'")
        
    def load(self):
        if '__promises_dict__' in self.__dict__ and self.__promises_dict__:
            for name in list(self.__promises_dict__):
                self.__getattr__(name)
            delattr(self, "__promises_dict__")

    @classmethod
    def from_dict(_cls, _dict:Dict[str, Any])->object:
        raise NotImplementedError

    def save(self, asynch:bool=False):
        if not hasattr(self, 'resource'):
            raise ResourceException('Unable save Savable, resource not found')
        resource = self.resource
        
        
        if hasattr(self, 'data_id') and self.data_id:
            res = resource.update(self, asynch)
            return res
        else:
            res = resource.create(self, asynch)
            return res


class TableData(EData):
    db:Resource
    data_id: UUID
    resource:Savable

    def __init__(self, data_id:object=None, resource:Resource=None, **kwargs) -> None:
        if data_id != None and isinstance(data_id, str) and is_uuid(data_id):
            data_id = UUID(data_id)

        if not isinstance(data_id, UUID):
            data_id = None

        super().__init__(data_id, resource)

    @property
    def table(self) -> Savable:
        return self.resource
    

    def __setattr__(self, name: str, value: Any) -> None:
        from .elist import Eset

        if not (hasattr(value, "__origin__") and value.__origin__ == Eset):
            return super().__setattr__(name, value)

        collection_epure = value.get_collection_epure(parent_obj=self, field_name=name)
        # value.collection_epure = collection_epure
        value._redefine_collection_epure(collection_epure)

        return super().__setattr__(name, value)
    
    # @property
    # def tp(self): # doesnt raise error when called
    #     raise AttributeError("property tp cannot be accessed outside of method decorated by @escript decorator")
    
    # @property
    # def dbp(self): # doesnt raise error when called
    #     raise AttributeError("property dbp cannot be accessed outside of method decorated by @escript decorator")

    @classmethod
    def from_dict_deep(_cls, _dict:Dict[str, Any])->object:

        # instance = _cls.__call__()
        instance = _cls()
        
        for field_name, val in _dict.items():
            if field_name in instance.annotations:
                _type = instance.annotations[field_name]
                if isinstance(_type, Savable):
                    val = _type(val)
                setattr(instance, field_name, val)


        return instance
    
    @classmethod
    def from_dict(cls, _dict:Dict[str, Any]) -> Epure:
        from .elist import ECollectionMetacls
        from .elist import Elist, Eset


        # instance = _cls.__call__()
        instance = cls()
        instance.__promises_dict__ = {}

        for field_name, val in _dict.items():

            if not field_name in instance.annotations:
                continue

            cls_attr_type = instance.annotations[field_name]

            if get_origin(cls_attr_type) != None: # check if class is Generic
                cls_attr_type = get_origin(cls_attr_type)
            
            val_type_match_cls_attr_type = isinstance(val, cls_attr_type)

            if field_name == "data_id" and not isinstance(val, UUID):
                val = UUID(val)
                val_type_match_cls_attr_type = True

            if isinstance(cls_attr_type, ECollectionMetacls) and not val_type_match_cls_attr_type: # check if attr is elist or eset
                # try:
                if is_uuid(val) and cls_attr_type.__origin__ == Elist:
                    eset_id = UUID(val)
                    promise = ElistPromise(cls_attr_type.collection_epure.resource, eset_id, cls_attr_type)
                    instance.__promises_dict__[field_name] = promise
                    continue

                elif is_uuid(val) and cls_attr_type.__origin__ == Eset:
                    eset_id = UUID(val)
                    promise = EsetPromise(cls_attr_type.collection_epure.resource, eset_id, cls_attr_type)
                    instance.__promises_dict__[field_name] = promise
                    continue

                elif type(val[0]) == dict and isinstance(cls_attr_type.py_type, Savable)\
                and cls_attr_type.__origin__ in (Elist, Eset):
                    val = val.copy()
                    for i, item in enumerate(val):
                        new_item = cls_attr_type.py_type()
                        for name, it in val[i].items():
                            setattr(new_item, name, it)
                        val[i] = new_item

                # except(Exception):
                elif cls_attr_type.py_type in (bytes, bytearray)\
                    and not type(val[0]) in (bytes, bytearray):
                        val = val.copy()
                        for i, item in enumerate(val):
                            val[i] = item.encode()

                val = cls_attr_type(val)

                val_type_match_cls_attr_type = True

            elif issubclass(cls_attr_type, Savable) and not val_type_match_cls_attr_type and is_uuid(val): # check if attr is epure 
                # try: 
                #     data_id = UUID(val)
                #     promise = DataPromise(cls_attr_type.resource, data_id)
                #     instance.__promises_dict__[field_name] = promise
                #     continue
                # except(Exception):
                #     val = cls_attr_type(val)
                promise = DataPromise(cls_attr_type.resource, val)
                instance.__promises_dict__[field_name] = promise
                continue

            elif issubclass(cls_attr_type, Savable) and not val_type_match_cls_attr_type and type(val) == dict:
                # data_id = val.get("data_id", None)
                val = cls_attr_type(**val)

            val_type_match_cls_attr_type = True
            
            if val is None:
                val_type_match_cls_attr_type = True

            if isinstance(val, str) and cls_attr_type in (bytes, bytearray): # bytearray is yet to be tested!
                val = cls_attr_type(val.encode())
                val_type_match_cls_attr_type = True

            if val_type_match_cls_attr_type:
                setattr(instance, field_name, val)
            else:
                raise TypeError(f'Value for field "{field_name}" with value "{val}" of type "{type(val)}" does not match expected attr type of class '\
                                 f'"{cls}" with value name "{field_name}" and type of attr "{cls_attr_type}"')


        return instance
    
    # def to_json(self) -> str:
    #     from ..db.table import NodePromise
    #     res = {}
    #     for field_name, field_type in self.annotations.items():
    #         if isinstance(field_type, Constraint):
    #             field_type = field_type.py_type
                
    #         if self.is_excluded(field_name, field_type):
    #             continue
    #         if field_name not in self.table.header:
    #             continue
    #         if not hasattr(self, field_name):
    #             continue

    #         field_val = getattr(self, field_name, None)

    #         if isinstance(field_val, NodePromise):
    #             field_val = getattr(field_val, 'node_id', None)
    #             # field_val = getattr(field_val, 'none2', None)

    #         #working for db:
    #         if field_val and isinstance(field_type, Savable)\
    #         and not isinstance(field_val, UUID):
    #             field_val = getattr(self, field_name, None)
    #             field_type = field_val.annotations['node_id']
    #             field_val = field_val.save(True).node_id

    #         if isinstance(field_val, UUID):
    #             field_val = str(field_val)

    #         res[field_name] = field_val

    #     jsonpickle.set_preferred_backend('json')
    #     jsonpickle.set_encoder_options('json', ensure_ascii=False)

    #     res = jsonpickle.encode(res)

    #     return res

    @classmethod
    def from_json(cls, json_str:str, decoder=jsonpickle.decode) -> Epure:

        decoder_res = decoder(json_str)

        res = cls.from_dict(decoder_res)

        return res


    def _serialize_field_val_to_dict(self, field_val, field_type=None, field_name:str=None, full:bool=None,_rec_depth:int = None, lambda_func:Callable=None):
        # return super()._serialize_field_val(field_type)
        # lambda_func = args[0][0]

        if isinstance(field_val, Savable) and not isinstance(type(field_val), ECollectionMetacls)\
        and (lambda_func(field_name, field_val, field_type, self, _rec_depth) or full == True):
            field_val = field_val.to_dict(full, lambda_func, _rec_depth+1)

        elif (isinstance(type(field_val), ECollectionMetacls) or type(field_val) in (set, tuple, list)) and len(field_val)\
        and isinstance(next(iter(field_val), False), Savable) and\
        (lambda_func(field_name, field_val, field_type, self, _rec_depth) or full == True):
            field_val = [item.to_dict(full, lambda_func, _rec_depth+1) for item in field_val]

        elif isinstance(type(field_val), ECollectionMetacls) and len(field_val) and not isinstance(next(iter(field_val), False), Savable)\
        and lambda_func(field_name, field_val, field_type, self, _rec_depth):
            field_val = [item for item in field_val]

        elif isinstance(field_val, Savable) and hasattr(field_val, "data_id"):
            field_val = field_val.data_id

        elif isinstance(field_val, Savable) and not hasattr(field_val, "data_id"):
            field_val = None

        if isinstance(field_val, UUID):
            field_val = str(field_val)

        return field_val


    def to_dict(self, full=False, lambda_func:Callable = lambda field_name, field_value, field_type, parent_value, rec_depth: 
                rec_depth < 1 or isinstance(type(parent_value), ECollectionMetacls), _rec_depth=0) -> Dict[str, Any]:

        self.load()

        _dict = self._serialize(self, self._serialize_field_val_to_dict, full, _rec_depth, lambda_func)
 
        return _dict

    def to_json(self, full=False, lambda_func:Callable = lambda field_name, field_value, field_type, parent_value, rec_depth: 
                rec_depth < 1 or isinstance(type(parent_value), ECollectionMetacls), _rec_depth=0, encoder:Callable=jsonpickle.encode) -> str:

        _dict = self.to_dict(full=full, lambda_func=lambda_func,_rec_depth=_rec_depth)
        
        return encoder(_dict)

    __exclude__:list = EData.__exclude__ + ['table', 'db']

class EsetTableData(TableData):

    def __hash__(self) -> int:
        val = self.value
        # if hasattr(val, "node_id") and val.node_id:
        #     return hash(val.node_id)
        
        return hash(id(val))


from __future__ import annotations
import types
from typing import TYPE_CHECKING
from typing import Any
from uuid import UUID
import collections.abc
# from ...epure import Epure

if TYPE_CHECKING:
    from .model import Model

class Term:

    def is_str_or_uuid(self, other:Any) -> bool:
        if isinstance(other,str) and "SELECT" not in other\
            or isinstance(other, UUID):
            return True

        return False
    
    def convert_tuple_for_in(self, other):
        other_types_set = {type(it) for it in other}
        if other_types_set.issubset({UUID, str}):
            other = [str(it) for it in other]
        other = tuple(other)
        return other

    #bool_ops
    def _and(self, left, right): #and
        return self.__db__.serialize_op('and', str(left), str(right))
    
    def _or(self, left, right): #or
        return self.__db__.serialize_op('or', str(left), str(right))
    
    #compare_ops
    def _not(self, other): #not
        return self.__db__.serialize_op('not', str(other))

    def not_in(self, other): #not in
        if type(other) in (set, list, tuple):
            other = self.convert_tuple_for_in(other)
        
        if isinstance(other, str) and not "(SELECT" in other\
            or not isinstance(other, tuple):
            raise TypeError(f"'{other}' is of type '{type(other)}' and not of type List, Tuple, Set or select method, so it cannot be right operand for SQL 'NOT IN' operator")
        
        return self.__db__.serialize_op('not in', str(self), str(other))
    
    def _in(self, other): #in
        if type(other) in (set, list, tuple):
            other = self.convert_tuple_for_in(other)
        
        if isinstance(other, str) and not "(SELECT" in other\
            or type(other) not in (tuple, str):
            raise TypeError(f"'{other}' is of type '{type(other)}' and not of type List, Tuple, Set or Model.select() method, so it cannot be right operand for SQL 'IN' operator")
            
        return self.__db__.serialize_op('in', str(self), str(other))

    def __eq__(self, other): # ==
        if self.is_str_or_uuid(other):
            other = repr(str(other))
        return self.__db__.serialize_op('==', str(self), str(other))
    
    def __lt__(self, other): #<
        if self.is_str_or_uuid(other):
            other = repr(str(other))
        return self.__db__.serialize_op('<', str(self), str(other))

    def __le__(self, other): #<=
        if self.is_str_or_uuid(other):
            other = repr(str(other))
        return self.__db__.serialize_op('<=', str(self), str(other))

    def __ne__(self, other): #!=
        if self.is_str_or_uuid(other):
            other = repr(str(other))
        return self.__db__.serialize_op('!=', str(self), str(other))

    def __gt__(self, other): #>
        if self.is_str_or_uuid(other):
            other = repr(str(other))
        return self.__db__.serialize_op('>', str(self), str(other))

    def __ge__(self, other): #>=
        if self.is_str_or_uuid(other):
            other = repr(str(other))
        return self.__db__.serialize_op('>=', str(self), str(other))
    
    def _is(self, other): # is
        if other != None:
            raise TypeError(f"operator IS can be used only with NULL value, but not with {other}")
        return self.__db__.serialize_op('is', str(self), "NULL")
        
    def is_not(self, other): # is not
        if other != None:
            raise TypeError(f"operator IS NOT can be used only with NULL value, but not with {other}")
        return self.__db__.serialize_op('is not', str(self), "NULL")
    
    #non-existent in python
    def like(self, other):
        if type(other) in (UUID, str):
            other = repr(str(other))
        return self.__db__.serialize_op('like', str(self), str(other))
    
    #all

    #any

    #between

    #some
    
    def __str__(self):
        return self.serialize(True)